/*
 * Schnupperstudium 2020
 *
 * Gedchtnisspiel
 *
 * 02.02.2020 Sven Tschirley   1. Release 4.2.2020
 *
 */

#include <msp430g2313.h>
#include <stdlib.h>

/****************************************************/

#define LED_R    BIT0
#define LED_G    BIT1
#define LED_B    BIT2
#define LED_Y    BIT3
#define LED_ALL  (LED_R|LED_G|LED_B|LED_Y)
#define BUTTON_R BIT4
#define BUTTON_G BIT5
#define BUTTON_B BIT6
#define BUTTON_Y BIT7
#define BUTTON_ALL (BUTTON_R|BUTTON_G|BUTTON_B|BUTTON_Y)

#define MAX_STEPS 16        // das kann man ja auch erhhen
#define MAX_FAILS 4	        // nach 4 mal verlieren und stepCtr wird
                            // Untigkeit attestiert
#define MAX_TIMEOUT 10

// Zustnde des Automaten
enum
  {
    MODE_GAME_START,
    MODE_GAME_SIMON,
    MODE_GAME_PLAYER,
    MODE_GAME_END_FAIL,
    MODE_GAME_END_WIN,
    MODE_GAME_SLEEP
  };

/****************************************************/
unsigned char steps[MAX_STEPS+1]; // Speicher
unsigned char currentStep = 0;    // gibt an, in welchem Schritt das Spiel ist
unsigned char stepCtr     = 0;    // Zhlt Spielereingaben
unsigned int  timeoutCtr  = 0;    // Zhlt fr Spieler-Timeout
unsigned char failCtr     = 0;    // Zhlt Verluste
unsigned char buttonIn    = 0;    // Zwischenspeicher fr Taste
unsigned char random;             // Zufallszahl zwischen 1 und 4
unsigned char idx;                // Zhlvariable
         int  delayDuration;      // Leuchtdauer
unsigned char gameState   = MODE_GAME_START; // Zustandsspeicher

volatile unsigned int lfsr = 0xACE1;
/****************************************************/
/*
 * Funktionsprototypen
 */

/****************************************************/
void          ConfigWDT( void );
void          ConfigClocks( void );
void          ConfigPorts( void );
void          ConfigTimerA0( void );
void          IlluminateLED( unsigned char color, unsigned int duration);
unsigned char getRandom( void );
void          delay_ms( int msec );

/*
 * Konfiguration des Watchdogs -- einfach ausschalten,
 * den brauchen wir nicht. Der Controller arbeitet auch
 * mit fast garkeiner Batteriespannung
 */
void ConfigWDT( void )
{
  WDTCTL = WDTPW + WDTHOLD; // Stop WDT
}

/*
 * Konfigurieren der Takterzeugung
 */
void ConfigClocks( void )
{
  BCSCTL1 = CALBC1_1MHZ; 		    // Set range
  DCOCTL = CALDCO_1MHZ;  		    // Set DCO step + modulation
  BCSCTL3 |= LFXT1S_2;                      // LFXT1 = VLO
  IFG1 &= ~OFIFG;                           // Clear OSCFault flag
  BCSCTL2 |= SELM_0 + DIVM_3 + DIVS_3;      // MCLK = DCO/8, SMCLK = DCO/8
}

/*
 * Konfigurieren des Port1
 * Alle LEDs sind Ausgnge
 * Alle Buttons sind Eingnge
 */
void ConfigPorts( void )
{
  P1DIR = 0;
  /* Initialisierung der LEDs                                   */
  P1DIR |= LED_ALL;     // Alle LED pins als Ausgang schalten
  P1OUT &= LED_ALL;  	// Alle LEDs ausschalten

  P1DIR &= ~BUTTON_ALL; // Buttons sind Eingaenge
  P1IE  |= BUTTON_ALL;  // P1.4 bis P1.7 interrupt enabled
  P1IES |= BUTTON_ALL;  // P1.4 bis p1.7 Hi/lo edge
  P1IFG &= ~BUTTON_ALL; // IFG cleared

  P2DIR |= BIT0;	    // Pin fr Scope-Debugging
}

/*
 * Konfigurieren des Timers fr Delay
 * Takt 1MHZ, Quelle ist SMCLK, CCR wird
 */
void ConfigTimerA0( void )
{
  TACCTL1 = OUTMOD_0;
  TACCR1  = 32;
  TACCTL0 = CCIE;            // Interrupt wird hier nicht gebraucht
  TACCR0  = 0x0100;          // setzt delay-Funktion neu
  TA0CTL  = TASSEL_2 + MC_2; // Source SMCLK, Cont.Mode
}

/****************************************************/
/*
 * Interrupt Service Routinen (ISR)
 */

/*
 * nicht verwentete Interrupts werden als leere Funktionen
 * implementiert, damit nicht irgendwas eingefangen wird.
 * Das sollte nich passieren, da die alle ausgeschaltet sind,
 * aber man kann ja nie wissen...
 */
#pragma vector=PORT1_VECTOR
__interrupt void Port_1_ISR( void ){
  P1IFG &= ~BUTTON_ALL;  // IFG lschen
  buttonIn = (~P1IN) & BUTTON_ALL; // Eingabe speichern
  delayDuration++;
}

//#pragma vector=ADC10_VECTOR
//__interrupt void ADC10_ISR( void ){}

#pragma vector=PORT2_VECTOR
__interrupt void Port_2_ISR( void ){}

#pragma vector=NMI_VECTOR
__interrupt void NMI_ISR( void ){}

#pragma vector=TIMER0_A0_VECTOR
__interrupt void Timer_A (void)
{
  // wake on CCR0 count match
  TA0CCTL0 = 0;
  //__bic_SR_register_on_exit(LPM0_bits|GIE);
}

#pragma vector=TIMER0_A1_VECTOR
__interrupt void Timer_A1_ISR( void ){}

//#pragma vector=USI_VECTOR
//__interrupt void USI_ISR( void ){}

#pragma vector=WDT_VECTOR
__interrupt void WDT_ISR( void ){}

/****************************************************/
/* Relevante Funktionen fr den Programmablauf	    */
/*
 * Einschalten einer LED fr eine bergebene Dauer
 */
void IlluminateLED( unsigned char color, unsigned int duration)
{
  P1OUT |= color;
  delay_ms( duration );
  P1OUT &= !color;
}

/*
 * Erzeugen einer Zufallszahl aus 1..4
 */
unsigned char getRandom( void )
{
  unsigned char tmp;
  tmp = ( rand() & 0x03 ) + 1;
  return( tmp );
}

/*
 * Timer-basierte Verzgerung
 * DCO ist 1 MHz, Timertakt 125kHz -> T = 8us
 * als CCR Wert 125, damit ca. 1 ms delay erzeugt wird,
 * so gut der interne Oszillator das kann
 */
void delay_ms(int  msec)
{
  int i = 0;
  // change according to your SMCLK frequency
  TA0CCR0 = 125;
  TA0CTL = TASSEL_2 + MC_2;
  for( i = 0; i < msec; ++i )
    {
      TA0R = 0;
      TA0CCTL0 = CCIE;
      while( TA0R < 80 )
        {
          __no_operation();
        }
    }
  TA0CTL = 0;
}

/****************************************************/
/*
 * Hauptprogramm
 */
int main ( void )
{
  /* Aufruf der Initialisierungsfunktion fuer den Controller    */
  ConfigWDT();		// Watchdog
  ConfigClocks();	// Takt
  ConfigPorts();	// LEDs
  ConfigTimerA0();	// Timer
  /* Zufallsgenerator initialisieren                            */
  srand( lfsr );

  // Zufall einmal laufenlassen
  for( idx = 0; idx<MAX_STEPS; idx++)
    {
	  steps[idx] = getRandom();
    }

  _BIS_SR( GIE );

  /* Endlosschleife mit Spiel                                   */
  while ( 1 ) // oder while( gameState != MODE_SLEEP )
    {
      switch( gameState )
        {
        case MODE_GAME_START:
          IlluminateLED( LED_R, 100 );
          IlluminateLED( LED_G, 100 );
          IlluminateLED( LED_B, 100 );
          IlluminateLED( LED_Y, 100 );
          IlluminateLED( LED_B, 100 );
          IlluminateLED( LED_G, 100 );
          IlluminateLED( LED_R, 100 );
          currentStep = 0;
          stepCtr     = 0;
          timeoutCtr  = 0;
          delayDuration = 500;
          delay_ms( 500 );		// Pause vor Start
          gameState = MODE_GAME_SIMON;
          break;
          // Abschnitt fr Ausgabe der Sequenz durch den Controller aus
          // dem Speicher bis zum erreichen von MAX_STEPS
        case MODE_GAME_SIMON:
          random = getRandom();
          // Zufallszahl der Farbe zuordnen und diese speichern
          if      ( random == 1 ) steps[currentStep++] = LED_R;
          else if ( random == 2 ) steps[currentStep++] = LED_G;
          else if ( random == 3 ) steps[currentStep++] = LED_B;
          else                    steps[currentStep++] = LED_Y;
          // Maximale Schrittzahl erreicht?
          if ( currentStep >= MAX_STEPS )
            {
              gameState = MODE_GAME_END_WIN;
            }
          // Inhalt des Speichers wiedergeben
          for ( idx=0; idx < currentStep; idx++ )
            {
              IlluminateLED( steps[ idx ], delayDuration );
              delay_ms( delayDuration );
            }
          // Zhler fr Spieler zurcksetzen
          stepCtr = 0;
          timeoutCtr = 0;
          delay_ms( 500 );       // Pause vor dem PLAYER
          // Spieler-Modus folgt, falls wir nicht schon gewonnen haben
          if ( gameState != MODE_GAME_END_WIN )
            {
        	  gameState = MODE_GAME_PLAYER;
            }
          break;
          // Spieler muss Sequenz wiedergeben
          //
        case MODE_GAME_PLAYER:
          while ( ( stepCtr < currentStep ) && ( gameState != MODE_GAME_END_FAIL ) )
            {
              // solange das Ende nicht erreicht wird luft diese Schleife,
              // die Eingaben werden in der ISR in buttonIn hineingeschrieben
              // die Durchlufe werden auf 500 ms verzgert
              if ( buttonIn != 0 )
                {
                  // Button RED
                  if ( buttonIn == BUTTON_R )
                    {
                      IlluminateLED( LED_R, 500 );
                      if ( steps[stepCtr] != LED_R )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  // Button GREEN
                  if ( buttonIn == BUTTON_G )
                    {
                      IlluminateLED( LED_G, 500 );
                      if ( steps[stepCtr] != LED_G )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  // Button BLUE
                  if ( buttonIn == BUTTON_B )
                    {
                      IlluminateLED( LED_B, 500 );
                      if ( steps[stepCtr] != LED_B )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  // Button YELLOW
                  if ( buttonIn == BUTTON_Y )
                    {
                      IlluminateLED( LED_Y, 500 );
                      if ( steps[stepCtr] != LED_Y )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  buttonIn = 0;
                  // nchsten Durchlauf vorbereiten
                  timeoutCtr = 0;
                  stepCtr++;
                } // if (buttonIn != 0 )
              else
                { // keine Eingabe, warte fr 5 sec, dann haben wir verloren
                  timeoutCtr++;
                  if ( timeoutCtr == MAX_TIMEOUT )
                    {
                      gameState = MODE_GAME_END_FAIL;
                    }
                  else
                    {
                      delay_ms( 500 );
                    }
                }
            } // while ( stepCtr < currentStep )
          delay_ms( 500 );   // Pause, dann ist SIMON dran
          // Spieler-Modus folgt, falls wir nicht schon verloren haben
          if ( gameState != MODE_GAME_END_FAIL )
            {
              gameState = MODE_GAME_SIMON;
            }
          break;
        case MODE_GAME_END_FAIL:
          IlluminateLED( LED_ALL, 1000 );
          delay_ms( 1000 );
          IlluminateLED( LED_ALL, 1000 );
          delay_ms( 1000 );
          if ( currentStep == 1 )
            {
        	  failCtr++;
            }
          if ( failCtr >= MAX_FAILS )
            {
        	  gameState = MODE_GAME_SLEEP;
        	  failCtr = 0;
            }
          else
            {
        	  gameState = MODE_GAME_START;
            }
          break;
        case MODE_GAME_END_WIN:
          for( idx=0; idx <= 5; idx++ )
            {
              IlluminateLED( LED_R, 100 );
              IlluminateLED( LED_B, 100 );
              IlluminateLED( LED_Y, 100 );
              IlluminateLED( LED_G, 100 );
              IlluminateLED( LED_Y, 100 );
              IlluminateLED( LED_G, 100 );
              IlluminateLED( LED_R, 100 );
              IlluminateLED( LED_B, 100 );
            }
          delay_ms( 1000 );
          gameState = MODE_GAME_START;
          break;
          // Schlafen bis Taste gedrckt wird, dann starten
        case MODE_GAME_SLEEP:
          P1OUT &= !LED_ALL;
          while ( buttonIn == 0 )  // wird von ISR beschrieben
            {
        	  delay_ms( 2000 );
        	  for ( idx=0; idx<4; idx++)
        	    {
        		  random = getRandom();
        	    }
        	  P2OUT |= BIT0;   // Lebenszeichen an P2.0
        	  delay_ms( 1 );
        	  P2OUT &= ~ BIT0;
            }
          if( buttonIn != 0 )
            {
        	  buttonIn = 0;
        	  gameState = MODE_GAME_START;
            }
        default:
          break;
        }
      //IlluminateLED( LED_B, 1000);
      _delay_cycles( 1000 );
    } // ende von  while ( 1 )
  return( 0 );

}

/* EOF */



